package jadex.adapter.jade;

import java.util.Map;

import jade.content.lang.Codec;
import jade.content.onto.Ontology;
import jade.core.Agent;
import jade.core.behaviours.Behaviour;
import jadex.adapter.*;
import jadex.adapter.fipa.AgentIdentifier;
import jadex.parser.SParser;
import jadex.util.DynamicURLClassLoader;
import jadex.util.collection.SCollection;
import jadex.config.Configuration;

/**
 *  The JADE wrapper agent is used to link the
 *  Jadex BDI runtime architecture to the JADE platform.
 *  The agent requires as first argument the agent model
 *  (MBDIAgent object, or filename of ADF).
 */
public class JadeAgentAdapter extends Agent implements IAgentAdapter
{
	//-------- constants --------

	/** The property name for specifying the max size of the message queue. */
	public static final String MESSAGE_QUEUE_MAX = "messagequeue.size";

	//-------- attributes --------

	/** The jadex agent. */
	protected IJadexAgent agent;

	/** The message receiver behaviour. */
	protected Behaviour mesrec;

	/** The timing behaviour. */
	protected TimingBehaviour timing;

	/** The agenda control behaviour. */
	protected ActionExecutionBehaviour agendacontrol;

	/** The agent thread (hack!!!). */
	protected transient	Thread	agentthread;

	/** The message event transport. */
	protected IMessageEventTransport[] transports;

	//-------- overridings --------

	/**
	 *  The setup method.
	 *  Automatically called when agents is born.
	 */
	protected void setup()
	{
		// super "constructor"
		super.setup();
		this.agentthread	= Thread.currentThread();
        // Use Jadex class loader as default. (hack???)
		// Might override jade specific settings (but unused in JADE 3.3)
        agentthread.setContextClassLoader(DynamicURLClassLoader.getInstance());

        // Set fallback configuration to JADE.
        Configuration.setFallbackConfiguration("jadex/adapter/jade/jade_conf.properties");
    	// Hack!!! Make platform name available to Jadex. (set always to override settings in properties)
    	String	name	= getName();
    	name	= name.substring(name.lastIndexOf("@")+1);
    	Configuration.getConfiguration().setProperty("platformname", name);

		// Initialize evaluation parameters, hide first argument (model).
		Object[]	args	= this.getArguments();
		if(args==null || args.length<2)
			throw new RuntimeException("Requires arguments (model, configuration).");
		Object[]	args2	= new Object[args.length-2];
		System.arraycopy(args, 2, args2, 0, args2.length);

		// Initialize the agent from model.
		if(args[0] instanceof String)
		{
			// Parse the arguments.
			Map argsmap = SCollection.createHashMap();
			for(int i=0; i<args2.length; i++)
			{
				// JADE agents can be supplied from command-line or via ams message with string arguments
				// They also can receive an object array via the inprocess interface (container controller)
				// Is this case the first argument must be a map that can be passed directly to Jadex.
				// Therefore arguments are interpreted when they are strings and contain a "=" character
				if(args2[i] instanceof String)
				{
					String tmp = (String)args2[i];
					int idx = tmp.indexOf("=");
					if(idx!=-1)
					{
						String argname = tmp.substring(0, idx);
						String argvalstr = tmp.substring(idx+1);
						//System.out.println("Found arg: "+argname+" = "+argvalstr);
						// Evaluate the argument value.
						Object argval = null;
						try
						{
							argval = SParser.evaluateExpression(argvalstr, null, null);
						}
						catch(Exception e)
						{
							System.out.println("Cannot evaluate argument: "+argname+". Reason: "+e.getMessage());
							//e.printStackTrace();
						}
						argsmap.put(argname, argval);
					}
				}
				else if(args2[i] instanceof Map)
				{
					argsmap.putAll((Map)args2[i]);
				}
			}
			
			this.agent = JadexAgentFactory.createJadexAgent(this, (String)args[0], (String)args[1], argsmap);
		}
//		else //if(args[0] instanceof IMBDIAgent)
//		{
//			this.agent = JadexAgentFactory.createJadexAgent(getName(), getLocalName(),
//				(IMBDIAgent)args[0], (String)args[1], params, this, "jadex.adapter.jade.jade_adapter");
//		}

		// Initialize Jade specific properties.
		String mmax = Configuration.getConfiguration().getProperty(MESSAGE_QUEUE_MAX);
		if(mmax==null)
		{
			agent.getLogger().warning("Could not read default message queue size from agent properties.");
		}
		else
		{
			int qs = Integer.parseInt(mmax);
			if(qs>0)
				setQueueSize(qs);
		}
		initLanguagesOntologies();

		this.transports = new IMessageEventTransport[1];
		transports[0] =	new JadeMessageEventTransport(this);

		// Start the required jade behaviours.
		// Hack!!! Add action execution behaviour first,
		// because it has to execute agent init action before other behaviours.
		this.agendacontrol	= new ActionExecutionBehaviour(agent);
		addBehaviour(agendacontrol);
		this.timing	= new TimingBehaviour(agent);
		addBehaviour(timing);
		this.mesrec	= new MessageReceiverBehaviour(agent);
		addBehaviour(mesrec);

		//try{Thread.sleep(20000);}catch(Exception e){}
		//System.out.println("Agent is starting: "+this.getName());
	}

	/**
	 *  Initialize Jade languages and ontologies.
	 *  Done on init and after agent move.
	 */
	// Todo: remove, replaced by content codec.
	protected void initLanguagesOntologies()
	{
		String[] names = agent.getPropertyNames("jade.language");
		for(int j=0; j<names.length; j++)
		{
			getContentManager().registerLanguage((Codec)agent.getProperty(names[j]));
		}
		names = agent.getPropertyNames("jade.ontology");
		for(int j=0; j<names.length; j++)
		{
			getContentManager().registerOntology((Ontology)agent.getProperty(names[j]));
		}
	}

	/**
	 *  Actions to perform when cloned.
	 */
	public void	afterClone()
	{
		initLanguagesOntologies();
		this.agentthread	= Thread.currentThread();
	}

	/**
	 *  Actions to perform when moved.
	 */
	public void	afterMove()
	{
		initLanguagesOntologies();
		this.agentthread	= Thread.currentThread();
	}

//	/**
//	 *  Takedown threads when threaded execution.
//	 */
//	protected void takeDown()
//	{
//		//System.out.println("Takedown called!!!");
//		// When jadex agent is still there, JADE agent is killed from outside.
//		// Perform cleanup.
//		if(agent!=null)
//		{
//			agent.cleanup();
//			agent	= null;
//		}
//	}

	//-------- IAgentAdapter methods --------

	/**
	 *  Notify the agent in a time interval.
	 *  @param millis The time in millis.
	 */
	public void notifyIn(long millis)
	{
		timing.notifyIn(millis);
	}

	/**
	 *  Test if the current thread is the agent thread.
	 *  @param thread The thread to test.
	 *  @return True, if thread is the agent thread.
	 */
	public boolean isAgentThread(Thread thread)
	{
		return thread==agentthread;
	}

	/**
	 *  Kill this agent.
	 */
	public void cleanupAgent()
	{
		agent	= null;
		super.doDelete();
	}

	/**
	 *  Kill this agent.
	 */
	public void doDelete()
	{
		// Add agenda action to avoid threading issues
		// super.doDelete() would interrupt agent thread (while e.g. waiting for plan).
		agent.killAgent();
	}

	/**
	 *  Called by the agent when it awoke from an idle state.
	 *  Platform has to make sure that agent will be executed
	 *  again from now on.
	 */
	public void	wakeup()
	{
		if(agendacontrol!=null)
			agendacontrol.restart();
	}

	/**
	 *  Used to determine the platform type.
	 *  Each adapter implementation should provide the value of a constant
	 *  defined somewhere.
	 *  Using the platform type, e.g. plan context conditions can check
	 *  if plans are applicable in current context. 
	 */
	public String getPlatformType()
	{
		return SJade.PLATFORM_TYPE;
	}

	/**
	 *  Return an agent-identifier that allows to send
	 *  messages to this agent.
	 */
	public AgentIdentifier getAgentIdentifier()
	{
		return SJade.convertAIDtoFipa(getAID());
	}

	/**
	 *  Get all avaiable transports available on the native platform.
	 *  @return All transports.
	 */
	public IMessageEventTransport[] getMessageEventTransports()
	{
		return transports;
	}
}
