package jadex.adapter.jade.tools.logger.gui;

import javax.swing.*;
import javax.swing.event.TableModelEvent;
import javax.swing.table.*;
import java.awt.*;
import java.util.*;
import jade.util.leap.ArrayList;
import jadex.adapter.jade.tools.logger.Logger;
import jadex.adapter.jade.tools.logger.ontology.PublishedLogRecord;
import jadex.util.jtable.*;

/**
 *
 */
public class LogTablePanel extends JPanel
{

    /** holds the strings to be displayed in the column headers of our table */
    final String[] columnNames = {"Sequence", "Date/Time", "Log-Level", "Agent", "Message", "Source-Class", "Source-Method", "Thread-ID"};

    /** holds the data for visible and hidden columns **/
    final boolean[] columnVisible = {true, false, true, true, true, true, true, false};


    protected JTable table;
    protected SorterFilterTableModel tableModel;
    protected DefaultTableModel dataModel;
    protected ResizeableTableHeader tableHeader;
    protected VisibilityTableColumnModel columnModel;


    private MainWindow mWnd;
    private MainPanel mPan;
    private PanelCanvas panCan; /* To resize and modify the scroll bars */
    private Logger myLogger;

    private boolean filterEnabled = false;
    private boolean autoscroll = false;


    public LogTablePanel(MainWindow mWnd, MainPanel mPan, Logger myLogger)
    {
        super();
        this.mWnd = mWnd;
        this.mPan = mPan;
        this.myLogger = myLogger;

        dataModel = new DefaultTableModel(columnNames, 0);
        tableModel = new SorterFilterTableModel(dataModel);
        columnModel = new VisibilityTableColumnModel();
        tableHeader = new ResizeableTableHeader(columnModel);

        table = new JTable(tableModel, columnModel)
        {
            public boolean isCellEditable(int row, int column)
            {
                return false;
            }

            public void tableChanged(TableModelEvent e)
            {
                super.tableChanged(e);
                if (e.getType() == TableModelEvent.INSERT && autoscroll)
                {
                    int row = e.getFirstRow();
                    // select new row
                    setRowSelectionInterval(row, row);
                    // scroll to new row
                    JViewport viewport = (JViewport) getParent();
                    Rectangle cellRect = getCellRect(row, -1, true);
                    Rectangle viewRect = viewport.getViewRect();
                    cellRect.setLocation(cellRect.x - viewRect.x, cellRect.y - viewRect.y);
                    cellRect.x = 0; // no horizontal scrolling
                    viewport.scrollRectToVisible(cellRect);
                }

            }

        };

        table.createDefaultColumnsFromModel();
        table.setAutoResizeMode(JTable.AUTO_RESIZE_OFF);
        table.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
        table.setTableHeader(tableHeader);


        // add sortrenderer
        tableModel.addMouseListener(table);
        // add column visibility popup
        columnModel.addMouseListener(table);


        for (int i = 0; i < columnModel.getAllColumnCount(); i++)
        {
            TableColumn column = columnModel.getAllColumn(i);
            boolean visible = columnVisible[i];
            columnModel.setColumnVisible(column, visible);
        }


    }

    public void addLogData(final PublishedLogRecord logData)
    {

        // ATTENTION!! In order to insert the received message in the queued message list
        // I cannot simply do it e.g.
        // ((DummyAgent)myAgent).getGui().queuedMsgListModel.add(0, (Object) new MsgIndication(msg, MsgIndication.INCOMING, new Date()));
        // WHY?
        // Because this is not thread safe!!!
        // In fact, if this operation is executed from this thread while the AWT REvent Dispatching
        // Thread is updating the JList component that shows the queued message list in the DummyAgent
        // GUI (e.g. because the user has just sent a message), this can cause an inconsistency
        // between what is shown in the GUI and what the queued message list actually contains.
        // HOW TO SOLVE THE PROBLEM?
        // I need to request the AWT REvent Dispatching Thread to insert the received message
        // in the queued message list!
        // This can be done by using the invokeLater static method of the SwingUtilities class
        // as below

        SwingUtilities.invokeLater(new EDTRequester(logData));

    }

    class EDTRequester implements Runnable
    {

        private PublishedLogRecord logData;

        public EDTRequester(PublishedLogRecord logData)
        {
            this.logData = logData;
        }

        public void run()
        {


            Vector row = new Vector();
            // append sequence number
            row.addElement(new Long(logData.getSequenceNumber()));
            // append milis as date
            Date date = new Date(new Long(logData.getMillis()).longValue());
            row.addElement(date);
            // append level
            row.addElement(logData.getLevel());
            // append loggername
            row.addElement(logData.getLoggerName());
            // append message
            row.addElement(logData.getMessage());
            // append source class name
            row.addElement(logData.getSourceClassName());
            // append source message name
            row.addElement(logData.getSourceMethodName());
            // append thread id
            row.addElement(new Integer(logData.getThreadID()).toString());

            //            dataModel.insertRow(0, row);
            dataModel.addRow(row);


            //            repaint();
        }

    }

    public void removeAllLogRecords()
    {
        dataModel.getDataVector().removeAllElements();
        dataModel.fireTableDataChanged();
    }


    public Vector getFilter()
    {
        return tableModel.getFilter();
    }

    public void setFilter(Vector filter)
    {
        tableModel.setFilter(filter);
    }

    public boolean isDoFilter()
    {
        return tableModel.isDoFilter();
    }

    public void setDoFilter(boolean filterEnabled)
    {
        tableModel.setDoFilter(filterEnabled);
    }

    public boolean isStopDisplaying()
    {
        return tableModel.isPausedDisplaying();
    }

    public void setStopDisplaying(boolean stopDisplaying)
    {
        tableModel.setPauseDisplaying(stopDisplaying);
    }

    public void refresh()
    {
        //        tableModel.changeRequest();
        tableModel.renewIndex();
    }

    public ArrayList getLogList()
    {
        int rows = tableModel.getRowCount();
        int cols = tableModel.getColumnCount();

        ArrayList al = new ArrayList();

        for (int row = 0; row < rows; row++)
        {
            StringBuffer line = new StringBuffer();
            for (int col=0; col<cols;col++) {
                line.append(tableModel.getValueAt(row,col).toString());
                line.append(";");
            }
            // delete last seperator
            line.deleteCharAt(line.length()-1);
            al.add(line.toString());
        }
        return al;
    }

    public void setAutoscroll(boolean autoscroll)
    {
        this.autoscroll = autoscroll;

    }

    public void resizeAllColumns()
    {
        ResizeableTableHeader tableHeader = (ResizeableTableHeader) table.getTableHeader();
        tableHeader.resizeAllColumns();
    }


}
